use omnipaxos::{
    ballot_leader_election::Ballot,
    storage::{Entry, Snapshot, StopSign, Storage, StorageResult},
    ClusterConfig,
};
use rand::Rng;
use serde::{Deserialize, Serialize};

#[derive(Debug, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct TestEntry(Vec<u64>);
#[derive(Debug, Clone, Serialize, Deserialize, PartialEq, Eq)]
pub struct TestSnapshot(Vec<TestEntry>);
impl Entry for TestEntry {
    type Snapshot = TestSnapshot;
}
impl Snapshot<TestEntry> for TestSnapshot {
    fn create(entries: &[TestEntry]) -> Self {
        Self(entries.to_vec())
    }
    fn merge(&mut self, delta: Self) {
        self.0.extend(delta.0)
    }
    fn use_snapshots() -> bool {
        true
    }
}

pub trait ValidateableStorage: Default {
    type Storage: Storage<TestEntry>;
    type Id;

    fn create(&mut self) -> (Self::Storage, Self::Id);
    fn close(&mut self, storage: Self::Storage, id: &Self::Id);
    fn open(&mut self, id: &Self::Id) -> Self::Storage;
    fn drop(&mut self, storage: Self::Storage, id: Self::Id);
}

fn random_entries(n: usize) -> Vec<TestEntry> {
    let mut rng = rand::thread_rng();
    (0..n).map(|_| TestEntry(vec![rng.gen()])).collect()
}

pub fn properties_kept_on_reopen(
    storage_factory: &mut impl ValidateableStorage,
) -> StorageResult<()> {
    let accepted_round = Ballot::with(42, 100, 3, 10);
    let compacted_idx = 192;
    let decided_idx = 42;
    let promise = Ballot::with(53, 23, 42, 623);
    let snapshot = Some(TestSnapshot(random_entries(4)));
    let stopsign = Some(StopSign::with(
        ClusterConfig::default(),
        Some(b"foo bar baz".to_vec()),
    ));

    let (mut storage, id) = storage_factory.create();
    storage.set_accepted_round(accepted_round)?;
    storage.set_compacted_idx(compacted_idx)?;
    storage.set_decided_idx(decided_idx)?;
    storage.set_promise(promise)?;
    storage.set_snapshot(snapshot.clone())?;
    storage.set_stopsign(stopsign.clone())?;

    storage_factory.close(storage, &id);
    let storage = storage_factory.open(&id);

    assert_eq!(
        &storage.get_accepted_round()?,
        &Some(accepted_round),
        "accepted_round kept over storage reopen"
    );
    assert_eq!(
        &storage.get_compacted_idx()?,
        &compacted_idx,
        "compacted_idx kept over storage reopen"
    );
    assert_eq!(
        &storage.get_decided_idx()?,
        &decided_idx,
        "decided_idx kept over storage reopen"
    );
    assert_eq!(
        &storage.get_promise()?,
        &Some(promise),
        "promise kept over storage reopen"
    );
    assert_eq!(
        &storage.get_snapshot()?,
        &snapshot,
        "snapshot kept over storage reopen"
    );
    assert_eq!(
        &storage.get_stopsign()?,
        &stopsign,
        "stopsign kept over storage reopen"
    );

    storage_factory.drop(storage, id);

    Ok(())
}

pub fn functionality_after_trim(
    storage_factory: &mut impl ValidateableStorage,
) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let entries = random_entries(6);

    storage.append_entries(entries.clone())?;
    assert_eq!(
        storage.get_log_len()?,
        6,
        "log has the correct length after first insert"
    );

    storage.trim(3)?;

    assert_eq!(
        storage.get_log_len()?,
        3,
        "log has the correct length after trim"
    );
    assert_eq!(
        storage.get_entries(3, 6)?,
        entries[3..6],
        "log has the correct entries after trim"
    );

    storage.append_on_prefix(5, entries.clone())?;
    assert_eq!(
        storage.get_log_len()?,
        8,
        "appending after trim appends the correct ammount"
    );
    assert_eq!(
        storage.get_entries(4, 7)?,
        [&entries[4..5], &entries[..2]].concat(),
        "appending after trim appends the correct entries"
    );

    assert_eq!(
        storage.get_suffix(5)?,
        entries,
        "get suffix after returns the correct items"
    );

    storage_factory.drop(storage, id);
    Ok(())
}

pub fn append_entries_reopen(storage_factory: &mut impl ValidateableStorage) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let entries = random_entries(16);
    storage.append_entries(entries.clone())?;
    assert_eq!(
        storage.get_log_len()?,
        entries.len(),
        "log has the correct length after first insert"
    );

    storage_factory.close(storage, &id);
    let storage = storage_factory.open(&id);

    assert_eq!(
        storage.get_entries(0, entries.len())?,
        entries,
        "appended entries"
    );

    storage_factory.drop(storage, id);
    Ok(())
}

pub fn append_on_prefix(storage_factory: &mut impl ValidateableStorage) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let entries = random_entries(6);

    storage.append_entries(entries.clone())?;
    storage.append_on_prefix(entries.len() / 2, entries.clone())?;

    assert_eq!(
        storage.get_log_len()?,
        9,
        "log has correct length after append_on_prefix"
    );

    assert_eq!(
        storage.get_entries(0, 9)?,
        [&entries[..3], &entries].concat(),
        "entries are appended at the correct spot"
    );

    storage_factory.drop(storage, id);
    Ok(())
}

pub fn get_suffix(storage_factory: &mut impl ValidateableStorage) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let entries = random_entries(6);

    storage.append_entries(entries.clone())?;

    assert_eq!(
        storage.get_suffix(3)?,
        &entries[3..6],
        "get suffix returns the correct items"
    );

    storage_factory.drop(storage, id);
    Ok(())
}

pub fn large_entry(storage_factory: &mut impl ValidateableStorage) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let mut rng = rand::thread_rng();
    let entry = TestEntry((0..1_000_000).map(|_| rng.gen()).collect());
    storage.append_entry(entry.clone())?;
    assert_eq!(
        storage.get_entries(0, 1)?,
        vec![entry],
        "large entry inserted and retrieved correctly"
    );

    storage_factory.drop(storage, id);
    Ok(())
}
pub fn many_entries(storage_factory: &mut impl ValidateableStorage) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let n = 1_000_000;

    let entries = random_entries(n);
    storage.append_entries(entries.clone())?;
    assert_eq!(
        storage.get_entries(0, n)?,
        entries,
        "many entries are correctly stored"
    );

    storage_factory.close(storage, &id);
    let storage = storage_factory.open(&id);

    assert_eq!(
        storage.get_entries(0, n)?,
        entries,
        "many entries are correctly retrieved after reopen"
    );

    storage_factory.drop(storage, id);
    Ok(())
}
pub fn large_stop_sign(storage_factory: &mut impl ValidateableStorage) -> StorageResult<()> {
    let (mut storage, id) = storage_factory.create();

    let mut rng = rand::thread_rng();
    let data = (0..1_000_000).map(|_| rng.gen()).collect();
    let stop_sign = Some(StopSign::with(ClusterConfig::default(), Some(data)));

    storage.set_stopsign(stop_sign.clone())?;

    assert_eq!(
        storage.get_stopsign()?,
        stop_sign,
        "large stop sign correctly stored"
    );

    storage_factory.close(storage, &id);
    let storage = storage_factory.open(&id);

    assert_eq!(
        storage.get_stopsign()?,
        stop_sign,
        "large stop sign correctly retrieved after reopen"
    );

    storage_factory.drop(storage, id);
    Ok(())
}

#[macro_export]
macro_rules! validate {
    (_TESTCASE $storage: ty, $function: ident) => {
        #[test]
        fn $function() {
            let mut storage: $storage = Default::default();
            $crate::$function(&mut storage).unwrap();
        }
    };

    ($storage: ty) => {
        $crate::validate!(_TESTCASE $storage, properties_kept_on_reopen);
        $crate::validate!(_TESTCASE $storage, functionality_after_trim);
        $crate::validate!(_TESTCASE $storage, append_entries_reopen);
        $crate::validate!(_TESTCASE $storage, append_on_prefix);
        $crate::validate!(_TESTCASE $storage, get_suffix);
        $crate::validate!(_TESTCASE $storage, large_entry);
        $crate::validate!(_TESTCASE $storage, many_entries);
        $crate::validate!(_TESTCASE $storage, large_stop_sign);
    };
}

#[cfg(test)]
mod tests {
    use super::*;

    use std::{
        collections::HashMap,
        path::{Path, PathBuf},
    };

    use omnipaxos_storage::{
        memory_storage::MemoryStorage,
        persistent_storage::{PersistentStorage, PersistentStorageConfig},
    };
    use tempfile::TempDir;

    mod memory {
        use super::*;

        #[derive(Default)]
        struct MemoryValidator {
            storages: HashMap<u64, MemoryStorage<TestEntry>>,
            next_id: u64,
        }

        impl ValidateableStorage for MemoryValidator {
            type Storage = MemoryStorage<TestEntry>;
            type Id = u64;

            fn create(&mut self) -> (Self::Storage, Self::Id) {
                let id = self.next_id;
                self.next_id += 1;
                (Self::Storage::default(), id)
            }
            fn close(&mut self, storage: Self::Storage, id: &Self::Id) {
                self.storages.insert(*id, storage);
            }
            fn open(&mut self, id: &Self::Id) -> Self::Storage {
                self.storages.remove(id).unwrap()
            }
            fn drop(&mut self, _storage: Self::Storage, _id: Self::Id) {}
        }

        validate!(MemoryValidator);
    }

    mod persistent {
        use super::*;

        fn config_for_path(path: &Path) -> PersistentStorageConfig {
            PersistentStorageConfig::with_path(path.to_str().unwrap().to_owned())
        }

        #[derive(Default)]
        struct PersistentValidator;

        impl ValidateableStorage for PersistentValidator {
            type Storage = PersistentStorage<TestEntry>;
            type Id = PathBuf;

            fn create(&mut self) -> (Self::Storage, Self::Id) {
                let path = TempDir::new().unwrap().into_path().join("store");

                (PersistentStorage::new(config_for_path(&path)), path)
            }
            fn close(&mut self, _storage: Self::Storage, _id: &Self::Id) {}
            fn open(&mut self, path: &Self::Id) -> Self::Storage {
                PersistentStorage::open(config_for_path(path))
            }
            fn drop(&mut self, _storage: Self::Storage, id: Self::Id) {
                let _ = std::fs::remove_dir_all(&id);
                let _ = std::fs::remove_dir(id.parent().unwrap());
            }
        }

        validate!(PersistentValidator);
    }
}
